#ifndef HARKTOOL4_STRINGVARIABLEIMPLANTER_HPP
#define HARKTOOL4_STRINGVARIABLEIMPLANTER_HPP

/** @file StringVariableImplanter.hpp
 * @brief BasicStringVariableImplanterクラス定義
 *
 * このファイルにはBasicStringVariableImplanterクラスと、そのtypedefである
 * StringVariableImplanter, WStringVariableImplanterの定義がある。
 *
 * StringVariableImplanterの概要
 * "{varname:format}"というパラメータ指定子を
 * boost::format(format) % varmap[varname]
 * で置換することができる。ここでvarmapというのは、文字列をキーとする連想配列
 * である。
 *
 * 例えば、"hello {var:%s}!"という文字列sと、{{"var","world"}}という連想配列m
 * があるとき、
 *   StringVariableImplanter svi(s);
 *   cout << (svi % m).str() << endl;
 * とすると"hello world!"と出力される。
 *
 * 連想配列に登録されていないキーは置換されないため、複数の連想配列を連続で
 * 使うこともできる。連想配列m0, m1, m2があるとき
 *   svi % m0 % m1 % m2;
 * などとできる。置換されずに残っているパラメータ指定子を削除したいときは
 * StringVariableImplanter::EraseParameterSpecifiers()
 * を使う。
 *   svi.EraseParameterSpecifiers();
 *   cout << svi.str() << endl;
 *
 * パラメータ指定子の記述に用いるブレース{と}を文字として表示したいときは
 * エスケープする。
 *   StringVariableImplanter svi("\\{{var\\}:%s}");
 * {{"var}","value"}}をmとすると、(svi % m).str() == "{value"
 *
 * パラメータ指定子の中に複数のコロン（:）を指定した場合、一番最後のコロン
 * が変数名とフォーマット文字列の区切りとして使われる。よって、変数名に
 * コロンが含まれていてもよい。
 *   StringVariableImplanter svi("{:var::%s}");
 * {{":var:","value"}}をmとすると、(svi % m).str() == "value"
 */
#include <string>
#include <map>
#include <utility>
#include <sstream>
#include <boost/regex.hpp>
#include <boost/optional.hpp>
#include <boost/format.hpp>

#include <iostream>

namespace harktool4
{

template <typename CharT, typename Traits = std::char_traits<CharT>, typename Allocator = std::allocator<CharT> >
class BasicStringVariableImplanter
{
public:
    typedef std::basic_string<CharT, Traits, Allocator> StringType;

    BasicStringVariableImplanter();

    explicit BasicStringVariableImplanter(const StringType& str);

    template <typename T>
    BasicStringVariableImplanter& operator %(const std::map<StringType, T>& variables);

    BasicStringVariableImplanter& EraseParameterSpecifiers();

    StringType str() const
    {
        return str_;
    }

    void str(const StringType& s)
    {
        str_ = s;
    }

private:
    StringType str_;

    struct ParamInfo
    {
        typename StringType::size_type start_pos_, colon_pos_, end_pos_;
    };

    /** @brief Returns a decoded string.
     * Deletes '\\' immediately anterior to a character.
     *
     * ex) if str == "foo\\Xbar" then the result is "fooXbar". '\\' anterior to 'X' was deleted.
     * ex) if str == "foobar\\" then the result is "foobar\\". '\\' at the end of str is not deleted.
     */
    static StringType DecodeEscapeSequence(const StringType& str);

    /** @brief Returns the position of c in this->str_.
     * Finds c and returns the position.
     * This function ignores c that is escaped with '\\'.
     *
     * ex) if str_ == "a\\bc" and c == 'b' then the result is StringType::npos.
     * ex) if str_ == "a\\bb" and c == 'b' then the result is 3.
     */
    const typename StringType::size_type FindNoEscape(CharT c, typename StringType::size_type pos = 0) const;

    const boost::optional<ParamInfo> SearchParamInfo(typename StringType::size_type pos = 0) const;

    const std::pair<StringType, StringType> ParamInfoString(const ParamInfo& param_info) const;

    template <typename T>
    void Implant(const std::map<StringType, T>& variables, bool erase_parameter_specifier);
};

template <typename CharT, typename Traits, typename Allocator>
BasicStringVariableImplanter<CharT, Traits, Allocator>::BasicStringVariableImplanter()
    : str_()
{}

template <typename CharT, typename Traits, typename Allocator>
BasicStringVariableImplanter<CharT, Traits, Allocator>::BasicStringVariableImplanter(const StringType& str)
    : str_(str)
{}

template <typename CharT, typename Traits, typename Allocator>
template <typename T>
BasicStringVariableImplanter<CharT, Traits, Allocator>&
    BasicStringVariableImplanter<CharT, Traits, Allocator>::operator %(const std::map<StringType, T>& variables)
{
    Implant(variables, false);
    return *this;
}

template <typename CharT, typename Traits, typename Allocator>
BasicStringVariableImplanter<CharT, Traits, Allocator>&
    BasicStringVariableImplanter<CharT, Traits, Allocator>::EraseParameterSpecifiers()
{
    std::map<StringType, int> dummy_variables;
    Implant(dummy_variables, true);
    return *this;
}

template <typename CharT, typename Traits, typename Allocator>
typename BasicStringVariableImplanter<CharT, Traits, Allocator>::StringType
    BasicStringVariableImplanter<CharT, Traits, Allocator>::DecodeEscapeSequence(const StringType& str)
{
    std::ostringstream oss;
    auto pos = str.find('\\');
    const auto len = str.length();

    oss << str.substr(0, pos);
    while (pos != StringType::npos && pos + 1 < len)
    {
        oss << str[pos + 1];
        pos += 2;
        const auto next_pos = str.find('\\', pos);
        if (next_pos == StringType::npos)
        {
            oss << str.substr(pos);
        }
        else
        {
            oss << str.substr(pos, next_pos - pos);
        }
        pos = next_pos;
    }

    return oss.str();
}

template <typename CharT, typename Traits, typename Allocator>
const typename BasicStringVariableImplanter<CharT, Traits, Allocator>::StringType::size_type
    BasicStringVariableImplanter<CharT, Traits, Allocator>::FindNoEscape(CharT c, typename StringType::size_type pos) const
{
    if (pos != StringType::npos)
    {
        // pos以降でcを検索
        pos = str_.find(c, pos);

        if (pos > 0)
        {
            // cが先頭ではない場合、エスケープシーケンスの可能性がある。
            while (pos != StringType::npos && str_[pos - 1] == '\\')
            {
                // エスケープシーケンス以外を発見するまでループ。
                pos = str_.find(c, pos + 1);
            }
        }
    }
    return pos;
}

template <typename CharT, typename Traits, typename Allocator>
const boost::optional<typename BasicStringVariableImplanter<CharT, Traits, Allocator>::ParamInfo>
    BasicStringVariableImplanter<CharT, Traits, Allocator>::SearchParamInfo(typename StringType::size_type pos) const
{
    typename StringType::size_type lbrace, rbrace, colon;

    // search {*:*}
    while (pos != StringType::npos)
    {
        pos = lbrace = FindNoEscape('{', pos);
        if (pos != StringType::npos)
        {
            pos = rbrace = FindNoEscape('}', lbrace + 1);
            if (pos != StringType::npos)
            {
                // found {*}
                colon = str_.rfind(':', rbrace - 1);
                if (colon != StringType::npos && lbrace < colon)
                {
                    // found {*:*}
                    return ParamInfo{ lbrace, colon, rbrace + 1 };
                }
                // not found, search next
                pos = rbrace + 1;
            }
        }
    }
    return boost::optional<ParamInfo>();
}

template <typename CharT, typename Traits, typename Allocator>
const std::pair<
    typename BasicStringVariableImplanter<CharT, Traits, Allocator>::StringType,
    typename BasicStringVariableImplanter<CharT, Traits, Allocator>::StringType>
    BasicStringVariableImplanter<CharT, Traits, Allocator>::ParamInfoString(const ParamInfo& param_info) const
{
    return std::pair<StringType, StringType>(
            DecodeEscapeSequence(str_.substr(param_info.start_pos_ + 1, param_info.colon_pos_ - param_info.start_pos_ - 1)),
            DecodeEscapeSequence(str_.substr(param_info.colon_pos_ + 1, param_info.end_pos_ - param_info.colon_pos_ - 2)));

}

template <typename CharT, typename Traits, typename Allocator>
template <typename T>
void BasicStringVariableImplanter<CharT, Traits, Allocator>::Implant(const std::map<StringType, T>& variables, bool erase_parameter_specifier)
{
    std::basic_ostringstream<CharT, Traits, Allocator> oss;
    typename StringType::size_type start = 0, end;

    auto param_info = SearchParamInfo();

    while (param_info)
    {
        // copy characters before a parameter specifier.
        end = param_info->start_pos_;
        oss << DecodeEscapeSequence(str_.substr(start, end - start));
        start = param_info->end_pos_;

        auto result = ParamInfoString(*param_info);
        auto it = variables.find(result.first);
        const auto& format_string = result.second;

        if (it != variables.end())
        {
            const auto& value = it->second;
            oss << boost::format(format_string) % value;
        }
        else if (!erase_parameter_specifier)
        {
            oss << str_.substr(param_info->start_pos_, param_info->end_pos_ - param_info->start_pos_);
        }
        param_info = SearchParamInfo(param_info->end_pos_);
    }

    // copy the rest of characters.
    oss << str_.substr(start);

    str_ = oss.str();
}

typedef BasicStringVariableImplanter<char> StringVariableImplanter;
typedef BasicStringVariableImplanter<wchar_t> WStringVariableImplanter;

}

#endif // HARKTOOL4_STRINGVARIABLEIMPLANTER_HPP

